import asyncio, aiohttp
import json
import traceback
from copy import deepcopy

import beeprint
from ensureTaskCanceled import ensureTaskCanceled
from loguru import logger

import websockets
from websockets import WebSocketClientProtocol
from NoLossAsyncGenerator import NoLossAsyncGenerator


class BinanceWs:
    restful_baseurl = 'https://api.binance.com'
    ws_baseurl = 'wss://stream.binance.com:9443'

    def __init__(self, apikey):
        self._apikey = apikey
        self._session: aiohttp.ClientSession = None
        self._ws: websockets.WebSocketClientProtocol = None
        self._ws_generator: NoLossAsyncGenerator = None
        self._ws_ok: asyncio.Future = None
        self._handlers = set()
        self._exiting = False

    async def exit(self):
        self._exiting = True
        session_close_task = None
        ws_close_task = None
        if self._session:
            session_close_task = asyncio.create_task(self._session.close())
        if self._ws:
            ws_close_task = asyncio.create_task(self._ws.close())
        if session_close_task:
            await session_close_task
        if ws_close_task:
            await ws_close_task

    # def _generate_signature(self, **kwargs):
    #     if 'recvWindow' not in kwargs.keys():
    #         kwargs['recvWindow'] = 5000
    #     if 'timestamp' not in kwargs.keys():
    #         kwargs['timestamp'] = int(datetime.datetime.utcnow().timestamp() * 1000)
    #     params = [(str(key) + '=' + str(value)) for key, value in kwargs.items()]
    #     msg = '&'.join(params).encode('utf-8')
    #     return hmac.new(self._secret.encode('utf-8'), msg, digestmod=hashlib.sha256).hexdigest()

    @property
    def session(self):
        if not self._session:
            self._session = aiohttp.ClientSession()

        return self._session

    async def _generate_listenkey(self):
        # ts = int(datetime.datetime.utcnow().timestamp() * 1000)
        async with self.session.post(
                self.restful_baseurl + '/api/v3/userDataStream',
                headers={'X-MBX-APIKEY': self._apikey},
                # data={
                #     'recvWindow': 5000,
                #     'timestamp': ts,
                #     'signature': self._generate_signature(recvWindow=5000, timestamp=ts)}
        ) as r:
            listenKey = (await r.json())['listenKey']
            print(listenKey)
            return listenKey

    # async def _extend_listenkey(self, new_listen_key: str):
    #     async with self.session.put(
    #             self.restful_baseurl + '/api/v3/userDataStream',
    #             headers={'X-MBX-APIKEY': self._apikey},
    #             data={'listenKey': new_listen_key}
    #     )as r:
    #         return await r.json()

    async def _close_listenkey(self, new_listen_key: str):
        async with self.session.delete(
                self.restful_baseurl + '/api/v3/userDataStream',
                headers={'X-MBX-APIKEY': self._apikey},
                data={'listenKey': new_listen_key}
        )as r:
            return await r.json()

    async def _infinity_post_listenKey(self):
        while True:
            await asyncio.create_task(asyncio.sleep(30 * 60))
            await self._generate_listenkey()

    async def _time_limitted_ws(self):
        self._ws = await websockets.connect(self.ws_baseurl + '/ws/' + await self._generate_listenkey())
        # 通知实例化完成
        if not self._ws_ok.done():
            self._ws_ok.set_result(None)
        # 传递值
        self._ws_generator = NoLossAsyncGenerator(self._ws)
        async for msg in self._ws_generator:
            msg = json.loads(msg)
            logger.debug('\n' + beeprint.pp(msg, output=False, string_break_enable=False, sort_keys=False))
            tasks = []
            for handler in self._handlers:
                if asyncio.iscoroutinefunction(handler):
                    tasks.append(asyncio.create_task(handler(deepcopy(msg))))
                else:
                    handler(deepcopy(msg))
            [await task for task in tasks]

    async def _ws_manager(self):
        # 半小时发一个ping
        asyncio.create_task(self._infinity_post_listenKey())
        while not self._exiting:
            # 20小时一换ws连接
            time_limitted_ws_task = asyncio.create_task(self._time_limitted_ws())

            async def sleep_then_raise():
                await asyncio.sleep(20 * 3600)
                raise TimeoutError('Time to change ws.')

            try:
                await asyncio.gather(time_limitted_ws_task, sleep_then_raise())
            except TimeoutError as e:  # 正常更换
                if str(e) == 'Time to change ws.' and isinstance(self._ws_generator, NoLossAsyncGenerator):
                    logger.debug('\n' + traceback.format_exc())
                else:  # 异常更换
                    logger.error('\n' + traceback.format_exc())
            except:  # 异常更换
                logger.error('\n' + traceback.format_exc())
            finally:
                if isinstance(self._ws, WebSocketClientProtocol):
                    # 等待可能累积的数据全部吐出来并关闭
                    await self._ws_generator.close()
                    asyncio.create_task(self._ws.close())
                    asyncio.create_task(ensureTaskCanceled(time_limitted_ws_task))

    @classmethod
    async def create_instance(cls, apikey):
        self = cls(apikey)
        self._ws_ok = asyncio.get_running_loop().create_future()
        # 启动ws管理器
        asyncio.create_task(self._ws_manager())
        await self._ws_ok
        return self

    def add_order_handler(self, handler):
        '''
        添加订单相关数据的处理函数或者异步函数，形如
        def handler(msg):
            ...
        或
        async def handler(msg):
            ...

        :param new_handler:
        :return:
        '''
        # 异步函数
        if asyncio.iscoroutinefunction(handler):
            async def new_handler(msg):
                if msg['e'] == "executionReport":
                    return await handler(msg)

            self._handlers.add(new_handler)
        else:
            def new_handler(msg):
                if msg['e'] == "executionReport":
                    return handler(msg)

            self._handlers.add(new_handler)

    def filter_stream(self, _filters: list = None):
        '''
        Filter the ws data stream and push the filtered data to the async generator which is returned by the method.
        Remember to explicitly call the close method of the async generator to close the stream.

        stream=binancews.filter_stream()

        #handle message in one coroutine:
        async for news in stream:
            ...
        #close the stream in another:
        close_task=asyncio.create_task(stream.close())
        ...
        await close_task


        :param _filters:
        :return:
        '''
        if _filters is None:
            _filters = []

        ag = NoLossAsyncGenerator(None)

        def handler(msg):
            if (_filters and any(
                    [all([((key in msg) and (value == msg[key])) for key, value in _filter.items()]) for _filter in
                     _filters])) \
                    or not _filters:
                ag.q.put_nowait(msg)

        self._handlers.add(handler)
        _close = ag.close

        async def close():
            self._handlers.remove(handler)
            await _close()

        ag.close = close
        return ag

    def order_stream(self):
        '''
        Filter the ws order data stream and push the filtered data to the async generator which is returned by the method.
        Remember to explicitly call the close method of the async generator to close the stream.


        stream=binancews.order_stream()

        #handle message in one coroutine:
        async for news in stream:
            ...
        #close the stream in another:
        close_task=asyncio.create_task(stream.close())
        ...
        await close_task

        :return:
        '''
        return self.filter_stream([{"e": "executionReport"}])


if __name__ == '__main__':
    pass



import json

import aiohttp
from AsyncWebsocketStreamInterface import AsyncWebsocketStreamInterface
import websockets
from loguru import logger


class BinanceFapiAsyncWs(AsyncWebsocketStreamInterface):
    ws_baseurl = 'wss://fstream.binance.com'
    restful_baseurl = 'https://fapi.binance.com'

    def __init__(self, apikey):
        super(BinanceFapiAsyncWs, self).__init__()
        self._apikey = apikey
        self._session: aiohttp.ClientSession = None
        self._delay_listenKey_invalid_running = False

    @property
    def session(self):
        if not self._session:
            self._session = aiohttp.ClientSession()

        return self._session

    async def _generate_listenkey(self, debug=False):
        if not self._delay_listenKey_invalid_running:  # 确保只运行一个心跳
            asyncio.create_task(self._delay_listenKey_invalid())
            self._delay_listenKey_invalid_running = True

        async with self.session.post(
                self.restful_baseurl + '/fapi/v1/listenKey',
                headers={'X-MBX-APIKEY': self._apikey},
                # data={
                #     'recvWindow': 5000,
                #     'timestamp': ts,
                #     'signature': self._generate_signature(recvWindow=5000, timestamp=ts)}
        ) as r:
            if not debug:
                listenKey = (await r.json())['listenKey']
                return listenKey
            else:
                return await r.json()

    async def _delay_listenKey_invalid(self):
        while True:
            await asyncio.create_task(asyncio.sleep(30 * 60))
            logger.debug('Time to delay listenKey invalid.')
            await self._generate_listenkey()

    async def _create_ws(self):
        ws = await websockets.connect(self.ws_baseurl + '/ws/' + await self._generate_listenkey())
        return ws

    async def _when2create_new_ws(self):
        listenKeyExpired_stream = self.stream_filter([{'e': 'listenKeyExpired'}])

        async def read_listenKeyExpired_stream(listenKeyExpired_stream):
            async for news in listenKeyExpired_stream:
                try:
                    return
                finally:
                    asyncio.create_task(listenKeyExpired_stream.close())

        read_listenKeyExpired_stream_task = asyncio.create_task(read_listenKeyExpired_stream(listenKeyExpired_stream))
        # 20小时更新连接一次，或者服务端推送消息listenKey过期
        await asyncio.create_task(
            asyncio.wait(
                [read_listenKeyExpired_stream_task, asyncio.sleep(20 * 3600)],
                return_when='FIRST_COMPLETED'))
        logger.debug('Time to update ws connection.')

    async def _parse_raw_data(self, raw_data):
        msg = json.loads(raw_data)
        return msg

    async def exit(self):
        super_exit_task = asyncio.create_task(super(BinanceFapiAsyncWs, self).exit())
        if self._session:
            await asyncio.create_task(self._session.close())
        await super_exit_task


if __name__ == '__main__':
    import signal
    import asyncio

    """
    信号值      符号      行为
    2          SIGINT    进程终端，CTRL+C
    9          SIGKILL   强制终端
    15         SIGTEM    请求中断
    20         SIGTOP    停止（挂起）进程 CRTL+D
    """


    def safely_exit():
        asyncio.create_task(safely_exit_management())


    loop = asyncio.get_event_loop()

    loop.add_signal_handler(signal.SIGTERM, safely_exit)
    loop.add_signal_handler(signal.SIGINT, safely_exit)


    async def safely_exit_management():
        bfws_task = asyncio.create_task(bfws.exit())
        await bfws_task
        print('Safely exit.')
        loop.stop()


    async def loop_task():
        global bfws
        bfws = BinanceFapiAsyncWs(input('apikey:'))
        stream = bfws.stream_filter()
        async for msg in stream:
            logger.info(msg)


    loop.create_task(loop_task())

    try:
        loop.run_forever()
    finally:
        loop.close()
    exit()
